// $Id: CommandProcessor.cpp,v 1.29 2001/08/04 16:56:54 cthuang Exp $
#include <algorithm>
#include "TclObject.h"
#include "CommandProcessor.h"

// This represents a Tcl command to execute.

struct Command {
    Tcl_ThreadId threadId;
    Tcl_Interp *interp;
    TclObject script;
    TclObject *pResult;
    int completionCode;
    Tcl_Mutex *pMutex;
    Tcl_Condition *pCondition;
    bool done;

    Command (
        Tcl_ThreadId p_threadId,
        Tcl_Interp *p_interp,
        TclObject p_script,
        TclObject *p_pResult,
        Tcl_Mutex *p_pMutex,
        Tcl_Condition *p_pCondition):
            threadId(p_threadId),
            interp(p_interp),
            script(p_script),
            pResult(p_pResult),
            pMutex(p_pMutex),
            pCondition(p_pCondition),
            done(false)
    { }

    // Execute the command.
    void execute();

    // Notify waiting threads that the command completed.
    void notify();
};

// This is the event used to send commands to other threads.

struct CommandEvent {
    Tcl_Event event;		/* Must be first */
    Command *pCommand;
};

struct ThreadSpecificData {
    Tcl_Mutex commandMutex;
    Tcl_Condition commandDone;
};

static Tcl_ThreadDataKey dataKey;


Singleton<CommandProcessor> CommandProcessor::ms_singleton;

CommandProcessor &
CommandProcessor::instance ()
{
    return ms_singleton.instance();
}

CommandProcessor::CommandProcessor ()
{
    m_asyncHandler = Tcl_AsyncCreate(
        asyncProc, reinterpret_cast<ClientData>(this));
}

CommandProcessor::~CommandProcessor ()
{
    // TODO: I would like to clean up resources here, but when the destructor
    // is called on static instances of this class, the Tcl memory subsystem
    // may have been finalized, making the argument point to an invalid memory
    // address.
//    Tcl_AsyncDelete(m_asyncHandler);
}

int
CommandProcessor::eval (Tcl_ThreadId threadId,
                        Tcl_Interp *interp,
                        TclObject script,
                        TclObject *pResult)
{
    ThreadSpecificData *tsdPtr = reinterpret_cast<ThreadSpecificData *>(
        Tcl_GetThreadData(&dataKey, sizeof(ThreadSpecificData)));
    Tcl_Mutex *pCommandMutex = &tsdPtr->commandMutex;
    Tcl_Condition *pCommandDone = &tsdPtr->commandDone;

    Command *pCommand = new Command(
        threadId,
        interp,
        script,
        pResult,
        pCommandMutex,
        pCommandDone);

    if (Tcl_GetCurrentThread() == threadId) {
        // If the concurrency model is apartment threaded, the thread calling
        // the server object method should be the same thread that called
        // CoInitialize, which should be the thread running the Tcl interpreter
        // in this case.
        pCommand->execute();

    } else {
        // Add command to the work queue.
        m_queue.push(pCommand);
        Tcl_AsyncMark(m_asyncHandler);

        if (pResult == 0) {
            return TCL_OK;
        }

        // Wait for the result.
        Tcl_MutexLock(pCommandMutex);
        while (!pCommand->done) {
            Tcl_ConditionWait(pCommandDone, pCommandMutex, NULL);
        }
        Tcl_MutexUnlock(pCommandMutex);
    }

    int completionCode = pCommand->completionCode;
    delete pCommand;
    return completionCode;
}

int
CommandProcessor::asyncProc (ClientData clientData,
                             Tcl_Interp * /*interp*/,
	                     int code)
{
    CommandProcessor *pCmdProc =
        reinterpret_cast<CommandProcessor *>(clientData);

    // Transfer command events to Tcl event queue.
    CommandProcessor::Queue &queue = pCmdProc->m_queue;
    while (!queue.empty()) {
        Command *pCommand = queue.front();

        CommandEvent *pEvent = reinterpret_cast<CommandEvent *>(
            Tcl_Alloc(sizeof(CommandEvent)));
        pEvent->event.proc = CommandProcessor::eventProc;
        pEvent->event.nextPtr = 0;
        pEvent->pCommand = pCommand;

        Tcl_ThreadQueueEvent(
            pCommand->threadId,
            reinterpret_cast<Tcl_Event *>(pEvent),
            TCL_QUEUE_TAIL);
        Tcl_ThreadAlert(pCommand->threadId);

        queue.pop();
    }

    return code;
}

// Handle the event in the target thread.
// Returns 1 to indicate that the event was processed.

int
CommandProcessor::eventProc (Tcl_Event *pEvent, int /*mask*/)
{
    CommandEvent *pCommandEvent = reinterpret_cast<CommandEvent *>(pEvent);
    Command *pCommand = pCommandEvent->pCommand;
    pCommand->execute();
    if (pCommand->pResult == 0) {
        delete pCommand;
    } else {
        pCommand->notify();
    }
    return 1;
}

void
Command::execute ()
{
    completionCode = Tcl_EvalObjEx(
        interp, script, TCL_EVAL_DIRECT | TCL_EVAL_GLOBAL);

    if (pResult != 0) {
        *pResult = Tcl_GetObjResult(interp);
    }
}

void
Command::notify ()
{
    Tcl_MutexLock(pMutex);
    done = true;
    Tcl_ConditionNotify(pCondition);
    Tcl_MutexUnlock(pMutex);
}
