// $Id: HandleSupport.cpp,v 1.9 2002/02/15 22:14:50 cthuang Exp $
#include <sstream>
#include "HandleSupport.h"

InternalRep::InternalRep (
    Tcl_Interp *interp,
    Tcl_ObjCmdProc *pCmdProc,
    ClientData objClientData):
	m_interp(interp),
        m_clientData(objClientData),
	m_refCount(0)
{
    std::ostringstream oss;
    oss << CmdNameType::name() << "0x" << std::hex << this;
    m_name = oss.str();

    m_command = Tcl_CreateObjCommand(
	m_interp,
	const_cast<char *>(m_name.c_str()),
	pCmdProc,
	objClientData,
	0);

    CmdNameType::instance().insert(m_name.c_str(), this);
}

InternalRep::~InternalRep ()
{
    CmdNameType::instance().erase(m_name.c_str());
    Tcl_DeleteCommandFromToken(m_interp, m_command);
}

void
InternalRep::addRef ()
{
    ++m_refCount;
}

long
InternalRep::release ()
{
    if (--m_refCount == 0) {
	delete this;
	return 0;
    }
    return m_refCount;
}


// This maps Tcl_Obj pointers to an internal representation.

class ObjToRepMap
{
    struct ThreadSpecificData
    {
        Tcl_HashTable m_hashTable;
        bool m_initialized;
    };

    static Tcl_ThreadDataKey m_threadDataKey;

    static void exitProc(ClientData);
    static Tcl_HashTable *hashTable();

public:
    static void insert(Tcl_Obj *pObj, InternalRep *pRep);
    static InternalRep *find(Tcl_Obj *pObj);
    static void erase(Tcl_Obj *pObj);
};

Tcl_ThreadDataKey ObjToRepMap::m_threadDataKey;

void
ObjToRepMap::exitProc (ClientData)
{
    Tcl_DeleteHashTable(hashTable());
}

Tcl_HashTable *
ObjToRepMap::hashTable ()
{
    ThreadSpecificData *tsdPtr = reinterpret_cast<ThreadSpecificData *>(
        Tcl_GetThreadData(&m_threadDataKey, sizeof(ThreadSpecificData)));
    if (!tsdPtr->m_initialized) {
        tsdPtr->m_initialized = true;
        Tcl_InitHashTable(&tsdPtr->m_hashTable, TCL_ONE_WORD_KEYS);
        Tcl_CreateThreadExitHandler(exitProc, 0);
    }
    return &tsdPtr->m_hashTable;
}

void
ObjToRepMap::insert (Tcl_Obj *pObj, InternalRep *pRep)
{
    int isNew;
    Tcl_HashEntry *pEntry = Tcl_CreateHashEntry(
        hashTable(), reinterpret_cast<char *>(pObj), &isNew);
    Tcl_SetHashValue(pEntry, pRep);
}

InternalRep *
ObjToRepMap::find (Tcl_Obj *pObj)
{
    Tcl_HashEntry *pEntry = Tcl_FindHashEntry(
        hashTable(), reinterpret_cast<char *>(pObj));
    if (pEntry == 0) {
        return 0;
    }
    return reinterpret_cast<InternalRep *>(Tcl_GetHashValue(pEntry));
}

void
ObjToRepMap::erase (Tcl_Obj *pObj)
{
    Tcl_HashEntry *pEntry = Tcl_FindHashEntry(
        hashTable(), reinterpret_cast<char *>(pObj));
    if (pEntry != 0) {
        Tcl_DeleteHashEntry(pEntry);
    }
}


const char CmdNameType::ms_name[] = "::tcom::handle";
Tcl_ObjType *CmdNameType::ms_pCmdNameType;
Tcl_ObjType CmdNameType::ms_oldCmdNameType;

Singleton<CmdNameType> CmdNameType::ms_singleton;

CmdNameType &
CmdNameType::instance ()
{
    return ms_singleton.instance();
}

CmdNameType::CmdNameType ()
{
    // Hijack Tcl's cmdName type.
    ms_pCmdNameType = Tcl_GetObjType("cmdName");
    ms_oldCmdNameType = *ms_pCmdNameType;
    ms_pCmdNameType->freeIntRepProc = freeInternalRep;
    ms_pCmdNameType->dupIntRepProc = dupInternalRep;
    ms_pCmdNameType->updateStringProc = updateString;
    ms_pCmdNameType->setFromAnyProc = setFromAny;
}

CmdNameType::~CmdNameType ()
{
    // Delete any left over application objects.
    m_map.forEach(Delete());

    // Restore original cmdName type.
    ms_pCmdNameType->freeIntRepProc = ms_oldCmdNameType.freeIntRepProc;
    ms_pCmdNameType->dupIntRepProc = ms_oldCmdNameType.dupIntRepProc;
    ms_pCmdNameType->updateStringProc = ms_oldCmdNameType.updateStringProc;
    ms_pCmdNameType->setFromAnyProc = ms_oldCmdNameType.setFromAnyProc;
}

void
CmdNameType::freeInternalRep (Tcl_Obj *pObj)
{
    InternalRep *pRep = ObjToRepMap::find(pObj);
    if (pRep != 0) {
	if (pRep->release() == 0) {
            ObjToRepMap::erase(pObj);
        }
    }

    ms_oldCmdNameType.freeIntRepProc(pObj);
}

void
CmdNameType::dupInternalRep (Tcl_Obj *pSrc, Tcl_Obj *pDup)
{
    // TODO: An object is duplicated if it is about to be modified.  We cannot
    // allow modifications on a handle.

    ms_oldCmdNameType.dupIntRepProc(pSrc, pDup);
}

// This should never be called because the string representation should already
// be valid.

void
CmdNameType::updateString (Tcl_Obj *pObj)
{
    ms_oldCmdNameType.updateStringProc(pObj);
}

int
CmdNameType::setFromAny (Tcl_Interp *interp, Tcl_Obj *pObj)
{
    // Check if the string represents an existing handle.
    InternalRep *pRep = CmdNameType::instance().find(pObj);
    if (pRep != 0) {
	pRep->addRef();
    }

    int result = ms_oldCmdNameType.setFromAnyProc(interp, pObj);

    if (result == TCL_OK && pRep != 0) {
        ObjToRepMap::insert(pObj, pRep);
    }
    return result;
}

Tcl_Obj *
CmdNameType::newObj (Tcl_Interp *interp, InternalRep *pRep)
{
    Tcl_Obj *pObj = Tcl_NewStringObj(const_cast<char *>(pRep->name()), -1);
    Tcl_ConvertToType(interp, pObj, ms_pCmdNameType);
    return pObj;
}
