// $Id: HandleSupport.h,v 1.23 2001/10/13 17:56:14 Administrator Exp $
#ifndef HANDLESUPPORT_H
#define HANDLESUPPORT_H

#include <tcl.h>
#include <string>
#include "tcomApi.h"
#include "HashTable.h"
#include "Singleton.h"

// This class represents an association from a cmdName instance to an
// application object.

class TCOM_API InternalRep
{
protected:
    Tcl_Interp *m_interp;
    std::string m_name;
    Tcl_Command m_command;
    ClientData m_clientData;
    long m_refCount;

public:
    InternalRep(
	Tcl_Interp *interp,
	Tcl_ObjCmdProc *pCmdProc,
	ClientData clientData);
    virtual ~InternalRep();

    // Get handle name.
    const char *name () const
    { return m_name.c_str(); }

    // Get pointer to the application object.
    ClientData clientData () const
    { return m_clientData; }

    void addRef();
    long release();
};

// This class extends InternalRep to associate with a specific application
// object class.  The class takes ownership of the passed in application object
// and is responsible for deleting it.

template<class AppType>
class AppInternalRep: public InternalRep
{
public:
    AppInternalRep (
	Tcl_Interp *interp,
	Tcl_ObjCmdProc *pCmdProc,
	AppType *pAppObject):
	    InternalRep(interp, pCmdProc, pAppObject)
    { }

    virtual ~AppInternalRep();
};

template<class AppType>
AppInternalRep<AppType>::~AppInternalRep ()
{
    delete reinterpret_cast<AppType *>(clientData());
}

// Handles are instances of Tcl's cmdName type which this class hijacks in
// order to map handles to application objects.

class TCOM_API CmdNameType
{
    // handle name prefix
    static const char ms_name[];

    // pointer to Tcl cmdName type
    static Tcl_ObjType *ms_pCmdNameType;

    // saved Tcl cmdName type
    static Tcl_ObjType ms_oldCmdNameType;

    // handle string representation to internal representation map
    typedef StringHashTable<InternalRep *> HandleStringToRepMap;
    HandleStringToRepMap m_map;

    // Tcl type functions
    static void freeInternalRep(Tcl_Obj *pObj);
    static void dupInternalRep(Tcl_Obj *pSrc, Tcl_Obj *pDup);
    static void updateString(Tcl_Obj *pObj);
    static int setFromAny(Tcl_Interp *interp, Tcl_Obj *pObj);

    friend class Singleton<CmdNameType>;
    static Singleton<CmdNameType> ms_singleton;

    CmdNameType();
    ~CmdNameType();

public:
    // Get instance of this class.
    static CmdNameType &instance();

    // Get prefix for handle names.
    static const char *name ()
    { return ms_name; }

    // Create handle.
    Tcl_Obj *newObj(Tcl_Interp *interp, InternalRep *pRep);
    
    // Insert handle to object mapping.
    void insert (const char *handleStr, InternalRep *pRep)
    { m_map.insert(handleStr, pRep); }

    // Get count of matching elements.
    size_t count (Tcl_Obj *pHandle) const
    { return (m_map.find(Tcl_GetStringFromObj(pHandle, 0)) == 0) ? 0 : 1; }

    // Get the object represented by the handle.
    InternalRep *find (Tcl_Obj *pHandle) const
    { return m_map.find(Tcl_GetStringFromObj(pHandle, 0)); }

    // Remove handle to object mapping.
    void erase (const char *handleStr)
    { m_map.erase(handleStr); }
};

// This class provides functions to map handles to objects of a specific
// application class.

template<class AppType>
class HandleSupport
{
    // Tcl command that implements the operations of the object
    Tcl_ObjCmdProc *m_pCmdProc;

public:
    HandleSupport (Tcl_ObjCmdProc *pCmdProc):
        m_pCmdProc(pCmdProc)
    { }

    // Create a handle and associate it with an application object.  Takes
    // ownership of the application object and is responsible for deleting it.
    Tcl_Obj *newObj(Tcl_Interp *interp, AppType *pAppObject);

    // Get count of matching elements.
    size_t count (Tcl_Obj *pHandle) const
    { return CmdNameType::instance().count(pHandle); }

    // Get the application object represented by the handle.  If the handle
    // is invalid, return 0.
    AppType *find(Tcl_Obj *pHandle) const;
};

template<class AppType>
Tcl_Obj *
HandleSupport<AppType>::newObj (Tcl_Interp *interp, AppType *pAppObject)
{
    AppInternalRep<AppType> *pRep = new AppInternalRep<AppType>(
	interp, m_pCmdProc, pAppObject);
    return CmdNameType::instance().newObj(interp, pRep);
}

template<class AppType>
AppType *
HandleSupport<AppType>::find (Tcl_Obj *pObj) const
{
    InternalRep *pRep = CmdNameType::instance().find(pObj);
    if (pRep == 0) {
	return 0;
    }
    return reinterpret_cast<AppType *>(pRep->clientData());
}

#endif
